#! /bin/bash

RED_COLOR="\E[31m"
GREEN_COLOR="\E[32m"
RESET_COLOR="\E[37m"

# store all arguments in ARGS. 
ARGV=("$@")
ARGC=${#ARGV[@]}

# the first argument is the representation matrix of the problem.
ROOT=${ARGV[0]}
DOC="${ROOT}/doc"
PLOT_FOLDER=${DOC}/gnuplot

BINARY=${ARGV[1]}
REFERENCE_PROGRAM=${ARGV[2]}

PLOT_FILE="${DOC}/performance.dat"
CONFIGURATION=( "bp-trees" "bp-qbn-trees" "vp-trees" ) 

declare -a OVERALL_TIME
PROBLEM_COUNTER=1

if [ ! -e ${DOC} ]; then
    mkdir ${DOC}
fi
if [ ! -e ${PLOT_FOLDER} ]; then
    mkdir ${PLOT_FOLDER}
fi

if [ -e ${PLOT_FILE} ]; then
    rm ${PLOT_FILE}
fi

function traverse () {
    local FOLDER=$1

    for FILE in `ls -Sr ${FOLDER}/* | egrep '\.ine$|\.ver$'`; do
        if [ -d "$FILE" ]; then
            traverse "${FILE}"
        else
            if [ ${FILE: -4} == ".ine" ] || [ ${FILE: -4} == ".ver" ]; then
                TABLE="${TABLE}\t\t<tr>"
                run "${FILE}"
                TABLE="${TABLE}</tr>\n"
            fi
        fi
    done
}

function run () {
    local PROBLEM=$1
    local DESCRIPTION=`basename ${PROBLEM} .ine`

    TABLE="${TABLE}<td>${DESCRIPTION}</td>"
    OUTPUT_FILE_SIZE=0

    if [ ! -z ${REFERENCE_PROGRAM} ]; then
        echo -n "Running reference check for ${DESCRIPTION} ..."
        `./${REFERENCE_PROGRAM} ${PROBLEM} --avisfukudaformat >> /dev/null`
        mv ${PROBLEM}.out ${PROBLEM}.ref
        echo -e "${GREEN_COLOR}done.${RESET_COLOR}"
    fi

    for (( i = 0; i < ${#CONFIGURATION[@]}; i++ )); do
        STRATEGY=${CONFIGURATION[${i}]}

        local LOG="${DOC}/${DESCRIPTION}.${STRATEGY}.log"

        echo -n "Computing ${DESCRIPTION} for ${STRATEGY}..."

        ${BINARY} ${PROBLEM} settings_${STRATEGY}.conf > ${LOG}
        cp gnuplot.dat ${PLOT_FOLDER}/${DESCRIPTION}.${STRATEGY}.dat

        local TIME=`grep "Narrowing and verification time:" ${LOG} | cut -d' ' -f 5`
        local LOG_PATH=`basename ${LOG}`
        TABLE="${TABLE}<td><a href=\"${LOG_PATH}\">${TIME%??}</a></td>"

        OVERALL_TIME[${i}]=$((${OVERALL_TIME[${i}]}+${TIME%??}))

        if [ ! -z ${REFERENCE_PROGRAM} ]; then
            REF_DIFF=`diff -b <(sed 1d ${PROBLEM}.out | sort -n) <(sed 1,4d ${PROBLEM}.ref | sort -n)`
            if [ "${REF_DIFF}" != "" ]; then
                echo -e "${RED_COLOR}inconsistent output differs from reference!${RESET_COLOR}"
            fi
        fi

        if (( ${OUTPUT_FILE_SIZE} > 0 )); then
            if (( ${OUTPUT_FILE_SIZE} != `stat -c%s ${PROBLEM}.out` )); then
                echo -e "${RED_COLOR}inconsistent output file!${RESET_COLOR}"
            else
                echo -e "${GREEN_COLOR}done in ${TIME}.${RESET_COLOR}"
            fi
        else
            OUTPUT_FILE_SIZE=`stat -c%s ${PROBLEM}.out`
            echo -e "${GREEN_COLOR}done in ${TIME}.${RESET_COLOR}"
        fi
    done  

    echo "${DESCRIPTION} ${PROBLEM_COUNTER} ${OVERALL_TIME[@]}" >> ${PLOT_FILE}
    PROBLEM_COUNTER=$((${PROBLEM_COUNTER}+1))
}

TABLE="<head><style type=\"text/css\"><!--@import url(\"style.css\");--></style></head>\n"
TABLE="${TABLE}<body>\n<table id=\"results_table\">\n"
# Table head
TABLE="${TABLE}\t<thead>\n"
TABLE="${TABLE}\t\t<th></th>\n"
for (( i = 0; i < ${#CONFIGURATION[@]}; i++ )); do
    STRATEGY=${CONFIGURATION[${i}]}
    
    TABLE="${TABLE}\t\t<th>${STRATEGY}</th>\n"
    OVERALL_TIME[${i}]=0
done 
TABLE="${TABLE}\t</thead>\n"

traverse $ROOT

TABLE="${TABLE}\t\t<tr><td>Total:</td>"
for (( i = 0; i < ${#CONFIGURATION[@]}; i++ )); do
    TABLE="${TABLE}\t\t<td>${OVERALL_TIME[${i}]}</td>\n"
done 
TABLE="${TABLE}</tr>\n</table></body>"
echo -e ${TABLE} > "${DOC}/benchmark.html"
