/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_BIT_COUNT_VISITOR__
#define __DD_BIT_COUNT_VISITOR__

#include <vector>

#include "DDBitList.h"
#include "DDRayRefVisitor.h"

#include "DDRayReferenceContainer.h"

namespace double_description {

    class BitPopulationEntry {

        private:
        unsigned short mPosition;
        unsigned long mCount;

        public:
        BitPopulationEntry( const unsigned short pos ) :
            mPosition( pos ),
            mCount( 0 ) {}

        bool operator< ( const BitPopulationEntry & rhs ) const {
            return mCount > rhs.mCount;
        }

        void reset ( unsigned short pos ) {
            mCount = 0;
            mPosition = pos;
        }

        void inc () {
            mCount++;
        }

        unsigned short position () const {
            return mPosition;
        }

        unsigned long count () const {
            return mCount;
        }
    };

    class BitCountVisitor : public RayRefVisitor {

        private:
        std::vector<unsigned long> mStats;
        unsigned long mCapacity;

        std::vector<BitPopulationEntry> mPopulation;
        unsigned long mTotalSetBits;

        protected:
        bool hasFinished () const {
            return true;
        }

        public:
        BitCountVisitor( const unsigned short capacity ) :
            RayRefVisitor(),
            mCapacity( capacity ) {

            for ( unsigned short pos = 0; pos < capacity; pos++ ) {
                mPopulation.push_back( BitPopulationEntry( pos ) );
            }
        }

        virtual ~BitCountVisitor () {}

        void onStart () {
            RayRefVisitor::onStart();

            mStats.assign( mCapacity, 0 );

            for ( unsigned short pos = 0; pos < mPopulation.size(); pos++ ) {
                mPopulation[ pos ].reset( pos );
            }     
            mTotalSetBits = 0;      
        }

        void onVisit ( const struct RayReference & ref ) {
            
            for ( unsigned long bit = 0; bit < mPopulation.size(); bit++ ) {
                if ( ref.activeSet.isSet( bit ) ) {
                    mPopulation[ bit ].inc();
                    mStats[ bit ]++;
                }
            }
        }

        void onFinish () {
            RayRefVisitor::onFinish();
            
            for ( unsigned short pos = 0; pos < mPopulation.size(); pos++ ) {
                mTotalSetBits += mPopulation[ pos ].count();
            }     

            std::sort( mPopulation.begin(), mPopulation.end() );
        }

        const std::vector<unsigned long> stats () const {
            return mStats;
        }

        const std::vector<BitPopulationEntry> & population () const {
            return mPopulation;
        }

        double getTotalSetBits () const {
            return mTotalSetBits;
        }

        const BitList getMostCommonBits ( unsigned short percentage ) const {
            BitList result( BitList::getBlocks( mPopulation.size() ) );
            unsigned long current_set_bits = 0;

            for ( unsigned long i = 0; i < mPopulation.size() && 100 * current_set_bits / mTotalSetBits < percentage; i++ ) {
                result.setBit( mPopulation[ i ].position() );
                current_set_bits += mPopulation[ i ].count();
            }

            return result;
        }
    };
}

#endif // __DD_BIT_COUNT_VISITOR__
