/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_CALCULATION_VECTOR__
#define __DD_CALCULATION_VECTOR__

#include "DDVector.h"

#include <vector>
#include <string.h>

#include <gmp.h>
#include <gmpxx.h>

namespace double_description {

	class CalculationVector : public Vector {

		public:
		CalculationVector ( const unsigned long dimension ) :
			Vector( dimension ) {}

		CalculationVector ( const CalculationVector & other ) :
			Vector( other ) {}

		virtual ~CalculationVector () {}

		unsigned long getDimension () const {
			return data.size();
		}
		
		bool isNonZeroAt ( const unsigned long index ) const {
			return data[index] != 0;
		}

		unsigned long getNonZeroCount () const {
			unsigned long count = 0;

			for ( unsigned long index = 0; index < getDimension(); index++ ) {
				if ( isNonZeroAt( index ) ) {
					count++;
				}
			}
			return count;
		}

		const mpz_class& getValue ( const unsigned long index ) const {
			return data[index];
		}

		void setValue (	const unsigned long index, const mpz_class& value ) {
			data[index] = value;
		}

		void setValue (	const unsigned long index, const long value ) {
			data[index] = value;
		}

		void add ( const unsigned long index, const mpz_class& increment ) {
			data[index] += increment;
		}

		void add ( const unsigned long index, const CalculationVector& vector ) {
			add( index, vector[index] );
		}

		void add ( const unsigned long begin, const unsigned long end, const CalculationVector& vector ) {

			for ( unsigned long index = begin; index < end; index++ ) {
				add( index, vector[index] );
			} 
		}

		void divide ( const unsigned long index, const mpz_class& divisor ) {
			data[index] /= divisor;
		}

		void divide ( const unsigned long begin, const unsigned long end, const mpz_class& divisor ) {

			for ( unsigned long index = begin; index < end; index++ ) {
				divide( index, divisor );
			} 
		}

		void multiply ( const unsigned long index, const  mpz_class& scalar ) {
			data[index] *= scalar;
		}

		void multiply ( const unsigned long begin, const unsigned long end, const mpz_class& scalar ) {

			for ( unsigned long index = begin; index < end; index++ ) {
				multiply( index, scalar );
			} 
		}

		void subtract ( const unsigned long index, const  mpz_class& subtractor ) {
			data[index] -= subtractor;
		}

		void subtract ( const unsigned long index, const CalculationVector& vector ) {
			subtract( index, vector[index] );
		}

		void negate ( const unsigned long index ) {
			data[index] = -data[index];	
		}

		const mpz_class& operator[] ( const unsigned long index ) const {
			return getValue( index );
		}

		bool operator== ( const CalculationVector& vector ) const {

			if ( vector.getDimension() != getDimension() )
				return false;

			for ( unsigned long index = 0; index < getDimension(); index++ ) {
				if( vector[index] != getValue( index ))
					return false;
			}

			return true;
		} 

		void calcScalarProduct ( const CalculationVector& vector, mpz_class& result ) const {
			result = 0;

			for ( unsigned long index = 0; index < getDimension(); index++ ) {
				result += vector[index] * getValue( index );
			}		
		}

		const std::string toString () const;
	};
}

#endif // __DD_CALCULATION_VECTOR__
