/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <cstdlib>

#include "DDInput.h"
#include "DDUtilsFactory.h"
#include "DDLexSortCriteria.h"
#include "DDZeroSortCriteria.h"

#include "DDLogging.h"
#include "DDOptions.h"

using namespace double_description;

/*
Input::Input ( const char* file ) :
	mSystem ( getMatrixFromFile ( file ) ),
	transformator ( applyPositiveBaseTransformation ( mSystem ) ),
	transformed ( mSystem * transformator ) {}
*/

Input::Input ( const char* file ) :
	mSystem ( getMatrixFromFile ( file ) ),
	transformator ( Input::getIdentityMatrix( mSystem.getColumns() )),
	transformed ( mSystem ) {}

const Matrix Input::getMatrixFromFile ( const char* file ) const {
	std::ifstream input;
    std::string line;

	input.open( file );
    input.exceptions(std::ifstream::eofbit);
	
	if ( !input ) {
		throw "Can not open file!";
	}

    try {
    /* Search for the start of the matrix definition - 'begin'. */
        do {
            std::getline( input, line );
            Options::remove_comments( line, '*' );
            Options::trim( line );        
        } while( line != "begin" );

        /* Read the matrix. */
        Matrix system = getMatrixFromFile( input );

        /* Check whether the matrix was correctly bounded. */
        input >> line;

        if ( line != "end" ) {
            DD_OUTPUT( "Matrix file seems to be incorrect. 'end' was expected but something different was found: " << line )
        }

	    system.sortDescending( LexSortCriteria(), 0, system.getRows() );
	    findBasis( system );

        return system;
    } catch ( std::ifstream::failure e ) {
        return Matrix( 0, 0 );
    }
}

const Matrix Input::getMatrixFromFile ( std::ifstream & input ) const {
	unsigned long dimension, inequalities;
	long value;
    std::string type;

	input >> inequalities;
	input >> dimension;
    input >> type;

	std::vector<unsigned long> order( inequalities );

	for ( unsigned long row = 0; row < inequalities; row++ ) {
        order[ row ] = row;
	} 

	Matrix system ( inequalities, dimension );

	for ( unsigned long row = 0; row < inequalities; row++ ) {
	    for ( unsigned long column = 0; column < dimension; column++ ) {
	        input >> value;
			system.setValue( order[ row ], column, value );
		}
	}

	return system;
}

void Input::findBasis ( Matrix & system ) const {
	RankCalculator calculator( system.getRows(), system.getColumns() );
	
	BitList rows( BitList::getBlocks( system.getRows() ));
	rows.setBit( 0 );
    calculator.load( system, rows );

	for ( unsigned short row = 1; row < system.getColumns(); row++ ) {
		rows.setBit( row );

        calculator.addRow( system.row( row ) );
		bool linear_independent = calculator.hasMinimumRank( row + 1 );
		if ( !linear_independent ) {
			for ( unsigned short sec_row = system.getColumns(); sec_row < system.getRows() && !linear_independent; sec_row++ ) {
				system.swapRows( row, sec_row );
                calculator.overwriteLastRow( system.row( row ) );
                linear_independent = calculator.hasMinimumRank( row + 1 );
			}
		}
		if ( !linear_independent ) {
			DD_OUTPUT( "Cannot extract a base with rank " << ( row + 1 ) )
			exit( -1 );
		} 
	}
}

const Matrix Input::applyPositiveBaseTransformation ( const Matrix & matrix ) const {
	Matrix base = matrix.getSubMatrix( 0, matrix.getColumns() );
	
	return applyNemhauserWolsey( base );
}

const Matrix Input::applyNemhauserWolsey ( Matrix& matrix ) const {

	Matrix C = Input::getIdentityMatrix( matrix.getRows() );
	Matrix transformation = Input::getIdentityMatrix( matrix.getRows() );
	unsigned long row, column;

	for ( row = 0; row < std::min( matrix.getRows(), matrix.getColumns() ); row++ ) {
	
		for ( column = row + 1; column < matrix.getColumns(); column++ ) {

			if ( matrix.isNonZeroAt( row, column ) ) {
				UtilsFactory::CALC_UTILS.calcExendedGCD ( matrix[row][row], matrix[row][column] );

				C.setValue ( row, row, UtilsFactory::CALC_UTILS.getRegisterS() );
				C.setValue ( column, row, UtilsFactory::CALC_UTILS.getRegisterT() );
				C.setValue ( row, column, -matrix[row][column]/UtilsFactory::CALC_UTILS.getResult() );
				C.setValue ( column, column, matrix[row][row]/UtilsFactory::CALC_UTILS.getResult() );

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, row, 1 );
				C.setValue ( column, row, 0 );
				C.setValue ( row, column, 0 );
				C.setValue ( column, column, 1 );
			}
		}
		
		if ( matrix.getValue( row, row ) < 0 ) {
			
			C.setValue ( row, row, -1 );
			
			matrix *= C;
			transformation *= C;	

			/* Transform C back to identity matrix */
			C.setValue ( row, row, 1 );
		}

		for ( column = 0; column < row; column++ ) {

			if ( matrix.isNonZeroAt( row, column ) ) {

				C.setValue ( row, column, -( matrix[row][column]/matrix[row][row] + 1 ) );

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, column, 0 );
			}
		} 

		for ( column = 0; column < row; column++ ) {

			if ( matrix.isNonZeroAt( row, column ) ) {

				if ( matrix[row][column] % matrix[row][row] != 0 ) {
					C.setValue ( row, column, ( abs( matrix[row][column]/matrix[row][row]) + 1 ) );
				} else {
					C.setValue ( row, column, abs( matrix[row][column]/matrix[row][row] ) );
				}

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, column, 0 );
			}
		} 
	}

	for ( row = 0; row < std::min( matrix.getRows(), matrix.getColumns() ); row++ ) {

		matrix.setValue( row, row, 1 );

		for ( column = 0; column < row; column++ ) {			

			if ( matrix.isNonZeroAt( row, column ) ) {

				C.setValue( row, column, -matrix[row][column] );

				matrix *= C;
				transformation *= C;

				/* Transform C back to identity matrix */
				C.setValue ( row, column, 0 );
			}
		}
	}	

	return transformation;	
}
