/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_MATRIX__
#define __DD_MATRIX__

#include <vector>
#include <cstring>
#include <string>
#include <gmp.h>
#include <gmpxx.h>

#include "DDTweaks.h"

#include "DDBitList.h"
#include "DDVector.h"
#include "DDSortCriteria.h"
#include "DDRepresentationVector.h"
#include "DDCalculationVector.h"
#include "DDUtilsFactory.h"

namespace double_description {

	class Matrix {

		private:
		std::vector<CalculationVector> mData;
		std::vector<unsigned long> mMapping;

		protected:
		CalculationVector & getRow ( const unsigned long index ) {
			return mData[mMapping[index]];
		}

		public:
		Matrix ();
		Matrix ( const unsigned long rows, const unsigned long columns );
		Matrix ( const unsigned long rows, const unsigned long columns, const std::vector<unsigned long>& mapping );
		Matrix ( const Matrix & other );

		unsigned long getRows () const {
			return mData.size();
		}

		virtual unsigned long getColumns () const {
            if ( unlikely( mData.empty() )) {
                return 0;
            } else {
			    return mData[0].getDimension();
            }
		}

		void insert ( const CalculationVector& row );

		void insert ( const CalculationVector& row, const unsigned short pos ) {
			mMapping.insert( mMapping.begin() + pos, mData.size() );
			mData.push_back( row );
		}

		const CalculationVector & getRow ( const unsigned long index ) const {
			return mData[mMapping[index]];
		}

		const CalculationVector & row ( const unsigned long index ) const {
			return mData[mMapping[index]];
		}

		const CalculationVector& getLastRow () const {
			return getRow( getLastRowIndex() );
		}

		unsigned long getLastRowIndex () const {
			return getRows() - 1;
		}

		bool isNonZeroAt ( const unsigned long row, const unsigned long column ) const {
			return getRow( row ).isNonZeroAt( column );
		}

		void setValue ( const unsigned long row, const unsigned long column, const mpz_class& value ) {
			getRow( row ).setValue( column, value );
		}

		void setValue ( const unsigned long row, const unsigned long column, const long value ) {
			getRow( row ).setValue( column, value );
		}

		const mpz_class& getValue ( const unsigned long row, const unsigned long column ) const {
			return getRow( row ).getValue( column );
		}

		const CalculationVector& operator[] ( const unsigned long index ) const {
			return getRow( index );
		}

		void operator = ( const Matrix& matrix );

		void operator *= ( const long scalar ) {
			for ( unsigned long index = 0; index < getRows(); index++ ) {
				mData[index] *= scalar;
			}
		}

		void operator *= ( const mpz_class& scalar ) {
			for ( unsigned long index = 0; index < getRows(); index++ ) {
				mData[index] *= scalar;
			}
		}

		void operator *= ( const Matrix & rhs );

		Matrix operator * ( const Matrix & rhs ) const;

		RepresentationVector operator * ( const RepresentationVector* rhs ) const {

			RepresentationVector vector( rhs->getDimension() );

			for ( unsigned long row = 0; row < getRows(); row++ ) {

				UtilsFactory::CALC_UTILS.loadRegisterS( 0 );

				for ( unsigned long index = 0; index < vector.getDimension(); index++ ) {
					if ( rhs->isNonZeroAt( index ) && isNonZeroAt( row, index ) ) {
						UtilsFactory::CALC_UTILS.loadRegisterT( rhs->getValue( index ) );
						UtilsFactory::CALC_UTILS.multiplyRegisterT( getValue( row, index ) );

						UtilsFactory::CALC_UTILS.incrementRegisterS( UtilsFactory::CALC_UTILS.getRegisterT() );
					}
				}

				if ( UtilsFactory::CALC_UTILS.getRegisterS() != 0 ) {
					vector.setValue( row, UtilsFactory::CALC_UTILS.getRegisterS() );
				}
			}

			return vector;
		}

		Matrix getSubMatrix ( const unsigned long firstRow, const unsigned long lastRow ) const;

		void swapRows ( const unsigned long first, const unsigned long second ) {
			unsigned long swap_element = mMapping[first];
			mMapping[first] = mMapping[second];
			mMapping[second] = swap_element;
		}

		void sort ( const SortCriteria & criteria, unsigned short start, unsigned short end ) {
			sort( criteria, start, end, 1 );
		}		
		void sortDescending ( const SortCriteria & criteria, unsigned short start, unsigned short end ) {
			sort( criteria, start, end, -1 );
		}
		void sort ( const SortCriteria & criteria, unsigned short start, unsigned short end, int swap_trigger );

		void optimizeRowsOrder ();

		const std::string getMapping () const;
		const std::string toString () const;
	};
}

#endif // __DD_MATRIX__
