/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_SLE_SOLVER__
#define __DD_SLE_SOLVER__

#include "DDCalculationMatrix.h"
#include "DDLinearDependenceEntry.h"

#include "DDLogging.h"

namespace double_description {

	class SLESolver : public CalculationMatrix {

		private:
		unsigned long mLastRowIndex;

		BitList mRowMap;

		protected:
		const unsigned long mBlocks;

		public:
		SLESolver ( const unsigned long rows, const unsigned long columns, const unsigned long blocks );
		~SLESolver () {}

		void load ( const Matrix& matrix, const BitList& rowMap ) {

			unsigned long last_column = 0;

			for ( unsigned long block = 0, prefix = 0; block < mBlocks; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = rowMap[block];

				while ( block_data ) {
					unsigned long suffix = __builtin_ctzl( block_data ); 
					unsigned long row = prefix + suffix;

					for ( unsigned long column = 0; column < matrix.getColumns(); column++ ) {
						setValue( column, last_column, matrix[row][column] );
					}
					last_column++;

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

			mRowMap = rowMap;
			setRows( matrix.getColumns() );
			setColumns( last_column );
		}

		void load ( const Matrix& matrix, const unsigned long target ) {

			unsigned long last_column = getColumns();

			for ( unsigned long column = 0; column < matrix.getColumns(); column++ ) {
				setValue( column, last_column, matrix[target][column] );
			}

			mLastRowIndex = target;
			setColumns( last_column + 1 );
		}
		
		void load ( const Matrix & matrix, const BitList & rowMap, const unsigned long target ) {

			unsigned long last_column;

			load( matrix, rowMap );			

			last_column = getColumns();

			for ( unsigned long column = 0; column < matrix.getColumns(); column++ ) {
				setValue( column, last_column, matrix[target][column] );
			}

			mLastRowIndex = target;
			setColumns( last_column + 1 );
		}

		void solve () {
			transformToUpperTriangularMatrix();
			transformToDiagonalMatrix();
			simplify();
		}

		void fillLinearDependenceEntry ( LinearDependenceEntry * const entry ) {
			unsigned long last_column = getColumns() - 1;

			entry->setScalingFactor( getValue( 0, 0 ) );

			for ( unsigned long block = 0, prefix = 0, row = 0; block < mBlocks; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = mRowMap[block];

				while ( block_data ) {
					unsigned long suffix = __builtin_ctzl( block_data ); 
					unsigned long index = prefix + suffix;

					if ( getRow( row ).isNonZeroAt( last_column ) ) {
						entry->insert( index, getValue( row, last_column ) );
					}
					row++;

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

			setColumns( last_column );
		}

		const LinearDependenceEntry * getLinearDependenceEntry () const {

			LinearDependenceEntry * result = new LinearDependenceEntry( mBlocks, mLastRowIndex + 1, getValue( 0, 0 ) );
			unsigned long last_column = getColumns() - 1;

			for ( unsigned long block = 0, prefix = 0, row = 0; block < mBlocks; block++, prefix += __BLOCK_SIZE__ ) {
				unsigned long block_data = mRowMap[block];

				while ( block_data ) {
					unsigned long suffix = __builtin_ctzl( block_data ); 
					unsigned long index = prefix + suffix;

					if ( getRow( row ).isNonZeroAt( last_column ) ) {
						result->insert( index, getValue( row, last_column ) );
					}
					row++;

					block_data ^= ( __SINGLE_BIT__ << suffix );
				}
			}

			return result;
		}
	};
}

#endif // __DD_SLE_SOLVER__
