/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_SAMPLES_CONTAINER__
#define __DD_SAMPLES_CONTAINER__ 

#include <string>
#include <sstream>

#include "DDBitList.h"

namespace double_description {

    class SamplesContainer {
        
        private:
        const unsigned int mRegularThreshold;
        const unsigned int mInverseThreshold;
        const unsigned int mCount;
        unsigned int mFilled;
        BitList * mSamples;

        static BitList * copySamples ( const BitList * samples, const unsigned int count ) {
            BitList * result = new BitList[ count ];

            for ( unsigned int index = 0; index < count; index++ ) {
                result[ index ] = samples[ index ];
            }

            return result;
        }

        public:
        SamplesContainer ( const unsigned int count,
                           const unsigned int regularThreshold,
                           const unsigned int inverseThreshold ) :
            mRegularThreshold( regularThreshold ),
            mInverseThreshold( inverseThreshold ),
            mCount( count ),
            mFilled( 0 ) {

            mSamples = new BitList[ 2*count ];
        }

        SamplesContainer ( const SamplesContainer & other ) :
            mRegularThreshold( other.mRegularThreshold ),
            mInverseThreshold( other.mInverseThreshold ),
            mCount( other.mCount ),
            mFilled( other.mFilled ),
            mSamples( copySamples( other.mSamples, 2*other.mCount ) ) {}

        ~SamplesContainer () {
            delete[] mSamples;
        }

        unsigned int getSamplesCount () const {
            return mFilled;
        }

        void insert ( const BitList & regular, const BitList & inverse ) {
            if ( mFilled < mCount ) {
                mSamples[ 2*mFilled ] = regular;
                mSamples[ 2*mFilled + 1] = inverse;
                mFilled++;
            }
        }
        
        void setFirst ( const BitList & regular, const BitList & inverse ) {

            if ( mCount > 0 ) {
                mSamples[0] = regular;
                mSamples[1] = inverse;
                if ( mFilled == 0 ) {
                    mFilled = 1;
                }
            }
        }

        void setLast ( const BitList & regular, const BitList & inverse ) {

            if ( mFilled > 0 ) {
                mSamples[ 2*(mFilled-1) ] = regular;
                mSamples[ 2*mFilled-1 ] = inverse;
            }
        }

        void remove () {
            mFilled--;
        }

        bool isRegularOver ( const BitList & map,
                             const unsigned int sample ) const {
        
            return map.getCommonPopCnt( getRegularSample( sample ) ) >= mRegularThreshold;
        }

        bool isRegularOver ( const BitList & map ) const {
        
            return isRegularOver( map, mFilled-1 );
        }

        bool isInverseOver ( const BitList & map,
                             const unsigned int sample ) const {
        
            return map.getCommonPopCnt( getInverseSample( sample ) ) >= mInverseThreshold;
        }

        bool isInverseOver ( const BitList & map ) const {
        
            return isInverseOver( map, mFilled-1 );
        }

        const BitList createInitialPattern () const {
            return BitList( BitList::getBlocks( mCount ), ULONG_MAX );
        }

        void refinePattern ( BitList & regPattern,
                             BitList & invPattern,
                             const BitList & map ) const {

            for ( unsigned int i = 0; i < mFilled; i++ ) {
                refinePattern( regPattern, invPattern, map, i );
            }
        } 

        void refinePattern ( BitList & regPattern,
                             BitList & invPattern,
                             const BitList & map,
                             const unsigned int sample ) const {

            if ( !isRegularOver( map, sample ) ) {
                regPattern.unsetBit( sample );
            } else if ( !isInverseOver( map, sample ) ) {
                invPattern.unsetBit( sample );
            }
        } 

        const BitList & getRegularSample ( const unsigned int index ) const {
            return mSamples[ 2*index ];
        }

        const BitList & getInverseSample ( const unsigned int index ) const {
            return mSamples[ 2*index+1 ];
        }        

        const std::string toString () const {
            std::stringstream output;

            output << "regular threshold: " << mRegularThreshold << std::endl
                   << "inverse threshold: " << mInverseThreshold << std::endl
                   << std::endl;

            for ( unsigned short index = 0; index < mFilled; index++ ) {
                output << index << " REG: " << getRegularSample( index ).toString() << std::endl
                       << index << " INV: " << getInverseSample( index ).toString() << std::endl
                       << std::endl;
            }

            return output.str();
        }
    };
}

#endif // __DD_SAMPLES_CONTAINER__ 
