/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_TREE_NODE__
#define __DD_TREE_NODE__

#include <vector>
#include <deque>

#include "DDLogging.h"

#include "DDTweaks.h"

#include "DDBitList.h"
#include "DDRayReference.h"
#include "DDPreliminaryCondition.h"
#include "DDExtVerificationCondition.h"

#include <string>
#include <fstream>

namespace double_description {

	class TreeNode {

		protected:
		const BitList mMap;
        const PreliminaryCondition *mCond;
		
		public:
		TreeNode ( const BitList & map ) : 
			mMap( map ),
            mCond( NULL ) {}

		TreeNode ( const BitList & map,
                   const PreliminaryCondition * const cond ) : 
			mMap( map ),
            mCond( cond ) {}

		virtual ~TreeNode () {
            if ( likely( mCond != NULL and mCond != &PreliminaryCondition::TRIVIAL_FAIL ) ) {
                delete mCond;
            }
        }

		const BitList & getMap () const { return mMap; }
        const PreliminaryCondition * getCondition () const { return mCond; }

		virtual void preliminaryTraverse ( const TreeNode * partner,
										    std::vector<struct EdgeCandidate> & candidates ) const = 0;

		virtual void preliminaryTraverse (  const RayReference & current,
									        std::deque<struct EdgeCandidate> & candidates,
                                            const TreeNode & verificator ) const = 0;

		virtual void preliminaryTraverse ( const RayReference * refs,
											const unsigned int refsCount,
											const PreliminaryCondition & condition,
											std::vector<struct EdgeCandidate> & candidates ) const = 0;

		virtual void preliminaryTraverse (  const RayReference & current,
									        std::deque<struct EdgeCandidate> & candidates ) const = 0;

        virtual bool verificationTraverse ( const ExtVerificationCondition & condition ) const = 0;

        virtual unsigned long getSubNodesCount () const = 0;
        virtual unsigned long getRaysCount () const = 0;
        virtual unsigned long getLeafNodesCount () const = 0;

        virtual bool isEmpty () const = 0;

        void toFile ( const std::string & fileName, const unsigned short depth ) const {

            std::fstream graph_file( fileName.c_str(), std::fstream::out );        

            graph_file << "digraph { " << std::endl
                       << toString( depth ) << std::endl
                       << "}";

            graph_file.close();
        }

        virtual void enumerateLeafNodes( std::vector<const TreeNode*> & container ) const = 0;

        virtual const std::string toString ( const unsigned short depth ) const = 0;
	};

    struct ComparatorMostBitsFirst {

        bool operator() ( const TreeNode * lhs, const TreeNode * rhs ) const {
            return lhs->getMap().getPopulationCount() > rhs->getMap().getPopulationCount();
        }
    };


    struct ComparatorMostRaysFirst {

        bool operator() ( const TreeNode * lhs, const TreeNode * rhs ) const {
            return lhs->getRaysCount() < rhs->getRaysCount();
        }
    };
}

#endif // __DD_TREE_NODE__
