/*  This is part of Addibit.
    Copyright (C) 2013 Blagoy Genov

    Addibit is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef __DD_VANTAGE_POINT_PRELIMINARY_CONDITION__
#define __DD_VANTAGE_POINT_PRELIMINARY_CONDITION__

#include "DDPreliminaryCondition.h"
#include "DDInterval.h"

namespace double_description {

	class VantagePointPreliminaryCondition : public PreliminaryCondition {

        protected:
    #ifndef __USE_TEMPLATES__
        const
    #endif
        BitList mVantagePoint;

        short mMinRegDistNarrowProceed;
        short mMinInvDistNarrowProceed;

        short mMaxRegDistVerProceed;
        short mMaxInvDistVerProceed;

        bool mUseVPForVerification;

		public:
    #ifdef __USE_TEMPLATES__
        VantagePointPreliminaryCondition () :
            PreliminaryCondition() {}

        void init( const BitList & point,
                   const Interval & distance,
                   const Interval & population,
                   const bool vp_verification ) {
            
            mVantagePoint = point;
            mUseVPForVerification = vp_verification;

            mMinRegDistNarrowProceed = getMinCommonBits() + distance.lb() - population.ub();
            mMinInvDistNarrowProceed = getMinCommonBits() - distance.ub();
            mMaxRegDistVerProceed = distance.ub();
            mMaxInvDistVerProceed = population.ub() - distance.lb();
        }
    #endif

		VantagePointPreliminaryCondition ( const BitList & map,
										   const unsigned int minCommonBits,
                                           const BitList & point ) :
            MapCondition( map ),
			PreliminaryCondition( map, minCommonBits ),
            mVantagePoint( point ),
            mUseVPForVerification( false ) {

                mMinRegDistNarrowProceed = 0;
                mMinInvDistNarrowProceed = 0;
                mMaxRegDistVerProceed = map.getPopulationCount();
                mMaxInvDistVerProceed = map.getPopulationCount();
        }

		VantagePointPreliminaryCondition ( const BitList & map,
										   const unsigned int minCommonBits,
										   const BitList & point,
                                           const Interval & distance,
                                           const Interval & population,
                                           const bool vp_verification ) :
            MapCondition( map ),
			PreliminaryCondition( map, minCommonBits ),
            mVantagePoint( point ),
            mUseVPForVerification( vp_verification ) {

                mMinRegDistNarrowProceed = getMinCommonBits() + distance.lb() - population.ub();
                mMinInvDistNarrowProceed = getMinCommonBits() - distance.ub();
                mMaxRegDistVerProceed = distance.ub();
                mMaxInvDistVerProceed = population.ub() - distance.lb();
        }

        virtual bool holdsOnFace ( const BitList & activeSetUnion ) const {
            return proceed_narrowing( activeSetUnion );
        }

        bool proceed_narrowing ( const BitList & set ) const {
            unsigned short inv_union_dist;
            const unsigned short reg_union_dist = mVantagePoint.getCommonPopCnt( set, inv_union_dist );

    #ifdef __STATS_OUTPUT__
            PreliminaryCondition::NARROWING_OPS++;
    #endif

            return reg_union_dist >= mMinRegDistNarrowProceed && inv_union_dist >= mMinInvDistNarrowProceed;
        }

        bool proceed_verification ( const BitList & set ) const {

            unsigned short inv_union_dist;
            const unsigned short reg_union_dist = mVantagePoint.getCommonPopCnt( set, inv_union_dist );

    #ifdef __STATS_OUTPUT__
            PreliminaryCondition::VERIFICATION_OPS++;
    #endif

            return reg_union_dist <= mMaxRegDistVerProceed &&
                   inv_union_dist <= mMaxInvDistVerProceed &&
                   ( reg_union_dist != mMaxRegDistVerProceed || inv_union_dist != mMaxInvDistVerProceed ) &&
                   PreliminaryCondition::proceed_verification( set );
        }

        bool isRedundant ( const VerificationCondition & cond ) const {
    #ifdef __STATS_OUTPUT__
            PreliminaryCondition::VERIFICATION_OPS++;
    #endif
            return mUseVPForVerification && cond.holdsOnRay( mVantagePoint );
        }

        const std::string toString () const {
            std::stringstream result;

            result << "min_n(d)=(" << mMinRegDistNarrowProceed << "," << mMinInvDistNarrowProceed << "), max_v=(" << mMaxRegDistVerProceed << "," << mMaxInvDistVerProceed << "), |v|=" << mVantagePoint.getPopulationCount();

            return PreliminaryCondition::toString() + ", " + result.str();            
        }
	};
}

#endif // __DD_VANTAGE_POINT_PRELIMINARY_CONDITION__
